@@ -21,7 +21,7 @@
 #ifndef NM_IP4_CONFIG_H
 #define NM_IP4_CONFIG_H
 
-#include <glib/gtypes.h>
+#include <glib.h>
 #include <glib-object.h>
 
 #include "nm-setting-ip4-config.h"
@@ -45,6 +45,7 @@ typedef struct {
 #define NM_IP4_CONFIG_NAMESERVERS "nameservers"
 #define NM_IP4_CONFIG_DOMAINS "domains"
 #define NM_IP4_CONFIG_ROUTES "routes"
+#define NM_IP4_CONFIG_WINS_SERVERS "wins-servers"
 
 GType nm_ip4_config_get_type (void);
 
@@ -67,6 +68,11 @@ guint32       nm_ip4_config_get_nameserver      (NMIP4Config *config, guint i);
 guint32       nm_ip4_config_get_num_nameservers (NMIP4Config *config);
 void          nm_ip4_config_reset_nameservers   (NMIP4Config *config);
 
+void          nm_ip4_config_add_wins            (NMIP4Config *config, guint32 wins);
+guint32       nm_ip4_config_get_wins            (NMIP4Config *config, guint i);
+guint32       nm_ip4_config_get_num_wins        (NMIP4Config *config);
+void          nm_ip4_config_reset_wins          (NMIP4Config *config);
+
 void          nm_ip4_config_take_route          (NMIP4Config *config, NMIP4Route *route);
 void          nm_ip4_config_add_route           (NMIP4Config *config, NMIP4Route *route);
 void          nm_ip4_config_replace_route       (NMIP4Config *config, guint32 i, NMIP4Route *new_route);
@@ -89,6 +95,9 @@ void          nm_ip4_config_set_mtu             (NMIP4Config *config, guint32 mt
 guint32       nm_ip4_config_get_mss             (NMIP4Config *config);
 void          nm_ip4_config_set_mss             (NMIP4Config *config, guint32 mss);
 
+gboolean      nm_ip4_config_get_never_default   (NMIP4Config *config);
+void          nm_ip4_config_set_never_default   (NMIP4Config *config, gboolean never_default);
+
 /* Flags for nm_ip4_config_to_rtnl_addr() */
 #define NM_RTNL_ADDR_NONE		0x0000
 #define NM_RTNL_ADDR_ADDR		0x0001
@@ -111,6 +120,7 @@ typedef enum {
 	NM_IP4_COMPARE_FLAG_SEARCHES    = 0x00000020,
 	NM_IP4_COMPARE_FLAG_MTU         = 0x00000040,
 	NM_IP4_COMPARE_FLAG_MSS         = 0x00000080,
+	NM_IP4_COMPARE_FLAG_WINS_SERVERS= 0x00000100,
 	NM_IP4_COMPARE_FLAG_ALL         = 0xFFFFFFFF   /* match everything */
 } NMIP4ConfigCompareFlags;
 
